/***************
 *
 * Note: when porting 32 bit apps to 64bit platform, remember _WIN32 is always defined on 64bit platform, 
 * _WIN64 is only defined on 64bit platform, so first check for _WIN64 then for _WIN32...
 *
 * #ifdef _WIN64
 *    win64 specific code
 * #elif defined(_WIN32)
 *    win32 specific code
 * #endif
 *
 **************/
#ifndef _UNICODE      /* Disable Unicode output from here */
#define _UNICODE
#endif

#ifdef _UNICODE
#ifndef UNICODE
#define UNICODE
#endif
#else
#ifdef UNICODE
#undef UNICODE
#endif
#endif

/* By default, compile for Windows XP */
#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0501
#endif

/* Assume IE 6 */
#ifndef _WIN32_IE
#define _WIN32_IE 0x0600
#endif

#include <windows.h>
#include <stdio.h>
#include <tchar.h>
#include <psapi.h>

typedef BOOL (WINAPI *LPFN_ISWOW64PROCESS) (HANDLE, PBOOL);

#ifdef _WIN64
#define PWORD DWORD64

typedef MEMORY_BASIC_INFORMATION64 GMEMORY_BASIC_INFORMATION;
typedef PMEMORY_BASIC_INFORMATION64 PGMEMORY_BASIC_INFORMATION;

#elif defined(_WIN32)
#define PWORD DWORD32

typedef MEMORY_BASIC_INFORMATION GMEMORY_BASIC_INFORMATION;
typedef PMEMORY_BASIC_INFORMATION PGMEMORY_BASIC_INFORMATION;

#endif

#define TERA_SHIFT 40
#define GIGA_SHIFT 30
#define MEGA_SHIFT 20
#define KILO_SHIFT 10

#define TERA    (((PWORD) 1 << TERA_SHIFT))
#define GIGA    ((PWORD)( 1 << GIGA_SHIFT))
#define MEGA    ((PWORD)( 1 << MEGA_SHIFT))
#define KILO    ((PWORD)( 1 << KILO_SHIFT))

#define GIGA_UNIT _T('G')
#define MEGA_UNIT _T('M')
#define KILO_UNIT _T('K')
#define TERA_UNIT _T('T')

#define PAGE_TRUNC(x, pagesize) ((x) & ~((pagesize) - 1))
#define __TFILE__ _T(__FILE__)
#define DumpError() __DumpError(GetLastError(), __TFILE__, __LINE__)
#define _CALC_PROT_(dwProt) (dwProt & ~(PAGE_GUARD | PAGE_NOCACHE))

static void PrintMemMap(HANDLE, LPVOID,LPVOID,DWORD);
static void __DumpError(DWORD, LPCTSTR, DWORD);
static void GetProtection(DWORD, LPTSTR);
static TCHAR GetState(DWORD);
static TCHAR GetType(DWORD);

static TCHAR* usageParaGraph =
  _T("This program dumps page map for the process.\r\n")
  _T("First Column contains address range.\r\n")
  _T("Second Column contains State of mapping. (C = Commit, F = Free, R = Reserved)\r\n")
  _T("Third Column contains Type of mapping. (I = Image, P = Private, M = Mapped, * = N/A)\r\n")
  _T("Fourth Column contains Type of protection (R = Read, W = Write, X = Execute, P = COW, G = Guard Page, ")
  _T("N = Not Cacheable, - = N/A, * = Unknown\r\n)")
  _T("Fifth Column (if available) contains full path to the file mapped\r\n")
  _T("\r\nNote: Process ID can be retrieved from Task Manager (taskmgr.exe)\r\n");

#if _WIN32_WINNT >= 0x0501

static LONG WINAPI ExHandler( PEXCEPTION_POINTERS);

static char oldOpCode;

#ifdef UNICODE
#define argv u_args
#else
#define argv args
#endif

int main(int argc, char** args)
{

  LPVOID pMain = 0;
  PWORD dwMain;
  DWORD dwPageSize;
  DWORD dwProcId;
  SYSTEM_INFO systemInfo;
  DWORD dwProt;
  TCHAR szOldProt[10];
  HANDLE hProc = 0;
  BOOL wow64Process = FALSE;
  LPFN_ISWOW64PROCESS fnIsWow64Process = 0;

#ifdef UNICODE

  LPWSTR lpCmdLine = NULL;
  LPWSTR* u_args = NULL;

  lpCmdLine = GetCommandLineW();
  u_args = CommandLineToArgvW(lpCmdLine, &argc);
#endif

  /* Hope kernel32.dll is loaded */
  hProc = GetModuleHandle(_T("kernel32"));
  fnIsWow64Process = (LPFN_ISWOW64PROCESS)GetProcAddress(hProc,"IsWow64Process");

  if(argc >= 2)
    dwProcId = _ttoi(argv[1]);
  else
    dwProcId = GetCurrentProcessId();

  hProc = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, dwProcId);
  if(!hProc)
    DumpError();

#ifdef _WIN64
  /* _tprintf(_T("64-bit build\n")); */
#elif defined(_WIN32)
  /*  _tprintf(_T("32-bit build\n")); */

  if(fnIsWow64Process)
	{
	  if((*fnIsWow64Process)(hProc, &wow64Process))
		{
		  if(wow64Process)
			_ftprintf(stderr, 
					  _T("Running under WoW64 subsystem, for accurate information use 64-bit version\n"));
		}
	}
#endif

  pMain = &GetProtection;
  dwMain = (PWORD)pMain;

  ZeroMemory(&systemInfo, sizeof(SYSTEM_INFO));
  ZeroMemory(&szOldProt, sizeof(TCHAR)*10);

  /*
   * Oops, for Win32 app running under WoW64 subsystem, the accurate system information
   * can be obtained by calling GetNativeSystemInfo() routine...
   *
   * but that fails this 32bit app under WoW64 (maximum virtual address)....
   */
  GetSystemInfo(&systemInfo);

  dwPageSize = systemInfo.dwPageSize;

  PrintMemMap(hProc, systemInfo.lpMinimumApplicationAddress,
			  systemInfo.lpMaximumApplicationAddress, systemInfo.dwPageSize);

  CloseHandle(hProc);

  if(dwProcId == GetCurrentProcessId())
    {
      _ftprintf(stderr, _T("\r\n\r\nUsage: %s [process-id]\r\n"), argv[0]);
      _ftprintf(stderr, usageParaGraph);
    }

#ifdef UNICODE
  if(u_args)
    GlobalFree(u_args);

#endif

#endif

  return 0;
}

#define PATHLEN (MAX_PATH << 2)

static void PrintMemMap(HANDLE hProc, LPVOID pMin, LPVOID pMax, DWORD dwPageSize)
{
  GMEMORY_BASIC_INFORMATION mem_info;
  PWORD i = ((PWORD)pMin);
  PWORD dwEnd;
  TCHAR chSize;
  TCHAR szProt[10];
  LPTSTR szModule = NULL;
  PWORD dwLen;
  HANDLE hHeap = GetProcessHeap();

  szModule = HeapAlloc(hHeap, HEAP_GENERATE_EXCEPTIONS | HEAP_ZERO_MEMORY, 
					   PATHLEN * sizeof(TCHAR));
  if(!szModule)
    DumpError();

  ZeroMemory(&mem_info, sizeof(GMEMORY_BASIC_INFORMATION));

  /*  _tprintf(_T("Maximum address is %p\n"), pMax); */

  while( i < ((PWORD)pMax))
    {
      if(!VirtualQueryEx(hProc, ((LPVOID)i), &mem_info, sizeof(GMEMORY_BASIC_INFORMATION)))
        DumpError();

      i += mem_info.RegionSize;

      if(!_CALC_PROT_(mem_info.Protect))
        continue;

      // Round off size
      dwEnd = (((PWORD)mem_info.BaseAddress) + mem_info.RegionSize);
#ifdef _WIN64
	  if(mem_info.RegionSize >= TERA)
		{
		  mem_info.RegionSize >>= TERA_SHIFT;
		  chSize = TERA_UNIT;
		}
      else 
#endif
		if(mem_info.RegionSize >= GIGA)
        {
          mem_info.RegionSize >>= GIGA_SHIFT;
          chSize = GIGA_UNIT;
        }
      else if(mem_info.RegionSize >= MEGA)
        {
          mem_info.RegionSize >>= MEGA_SHIFT;
          chSize = MEGA_UNIT;
        }
      else if(mem_info.RegionSize >= KILO)
        {
          mem_info.RegionSize >>= KILO_SHIFT;
          chSize = KILO_UNIT;
        }

      ZeroMemory(&szProt, sizeof(TCHAR)*10);
      GetProtection(mem_info.Protect, &szProt);

      _tprintf(
#ifdef _WIN64
			   _T("%016p-%016p")
#elif defined(_WIN32)
			   _T("%08p-%08p")
#endif
			   _T(" %c  %c  %6s  %5i%c"), mem_info.BaseAddress, 
			   dwEnd, GetState(mem_info.State), GetType(mem_info.Type), 
			   szProt, mem_info.RegionSize, chSize);
	  
      if(dwLen = GetMappedFileName(hProc, (void*)mem_info.BaseAddress,
                                   szModule, PATHLEN * sizeof(TCHAR)))
        {
          _tprintf(_T("  %s"), szModule);
          ZeroMemory(szModule, dwLen*sizeof(TCHAR));
        }
	  
      _tprintf(_T("\r\n"));
    }
  
  HeapFree(hHeap, 0, szModule);
}

static void __DumpError(DWORD dwErr, LPCTSTR szFile, DWORD dwLine)
{
  LPVOID lpMsgBuf = NULL;

  _ftprintf(stderr, _T("Error %i in %s at line no. %i.\r\n"), dwErr, 
			szFile, dwLine);
  if(FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
				   FORMAT_MESSAGE_FROM_SYSTEM | 
				   FORMAT_MESSAGE_IGNORE_INSERTS, NULL, dwErr,	
				   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                   (LPTSTR)&lpMsgBuf, 0, NULL))
    {
      _ftprintf(stderr, _T("Message: %s\r\n"), lpMsgBuf);
      LocalFree(lpMsgBuf);
    }
  ExitProcess(1);
}

static void GetProtection(DWORD dwProt, LPTSTR szProt)
{
  switch(_CALC_PROT_(dwProt))
    {
    case PAGE_EXECUTE:
      _tcsncpy(szProt, _T("--X-"), 4);
      break;
    case PAGE_EXECUTE_READ:
      _tcsncpy(szProt, _T("R-X-"), 4);
      break;
    case PAGE_EXECUTE_READWRITE:
      _tcsncpy(szProt, _T("RWX-"), 4);
      break;
    case PAGE_EXECUTE_WRITECOPY:
      _tcsncpy(szProt, _T("RWXP"), 4);
      break;
    case PAGE_NOACCESS:
      _tcsncpy(szProt, _T("----"), 4);
      break;
    case PAGE_READONLY:
      _tcsncpy(szProt, _T("R---"), 4);
      break;
    case PAGE_READWRITE:
      _tcsncpy(szProt, _T("RW--"), 4);
      break;
    case PAGE_WRITECOPY:
      _tcsncpy(szProt, _T("RW-P"), 4);
      break;
    default:
      _tcsncpy(szProt, _T("****"), 4);
    }

  if(dwProt & PAGE_GUARD)
    _tcsncat(szProt, _T("G"), 1);
  else
    _tcsncat(szProt, _T("-"), 1);

  if(dwProt & PAGE_NOCACHE)
    _tcsncat(szProt, _T("N"), 1);
  else
    _tcsncat(szProt, _T("-"), 1);
}

static TCHAR GetState(DWORD dwState)
{
  switch(dwState)
    {
    case MEM_COMMIT:
      return _T('C');
    case MEM_FREE:
      return _T('F');
    case MEM_RESERVE:
      return _T('R');
    default:
      return _T('*');
    }
}

static TCHAR GetType(DWORD dwType)
{
  switch(dwType)
    {
    case MEM_MAPPED:
      return _T('M');
    case MEM_IMAGE:
      return _T('I');
    case MEM_PRIVATE:
      return _T('P');
    default:
      return _T('*');
    }
}

/* Local Variables:
 * mode:java
 * mode:auto-fill
 * tab-width:4
 * End:
 */
/* vim:filetype=c
 */
